# -*- mode: python; coding: utf-8; -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""

.. testsetup ::

    from aiatools.aia import AIAFile

The :py:mod:`aiatools.aia` package provides the :py:class:`AIAFile` class for reading App Inventor (.aia) projects.
"""

import logging
import jprops
import os
from os.path import isdir, join
from zipfile import ZipFile
from .selectors import Selector, NamedCollection, UnionSelector
from .component_types import Screen

__author__ = 'Evan W. Patton <ewpatton@mit.edu>'

log = logging.getLogger(__name__)


class AIAAsset(object):
    """
    :py:class:`AIAAsset` provides an interface for reading the contents of assets from an App Inventor project.
    """

    def __init__(self, zipfile, name):
        """
        Constructs a new reference to an asset within an AIA file.

        :param zipfile: The zipped project file that is the source of the asset.
        :param name: The path of the asset within the project file hierarchy.
        """
        self.zipfile = zipfile
        self.name = name

    def open(self, mode='r'):
        """
        Opens the asset. ``mode`` is an optional file access mode.

        :param mode: The access mode to use when accessing the asset. Must be one of 'r', 'rU', or 'U'.
        :type mode: basestring
        :return: A file-like for accessing the contents of the asset.
        :rtype: zipfile.ZipExtFile
        """
        return self.zipfile.open(self.name, mode)


class AIAFile(object):
    """
    :py:class:`AIAFile` encapsulates an App Inventor project (AIA) file.

    Opens an App Inventor project (AIA) with the given filename. ``filename`` can be any file-like object that is
    acceptable to :py:class:`ZipFile`'s constructor, or a path to a directory containing an unzipped project.

    Parameters
    ----------
    filename : basestring | file
        A string or file-like containing the contents of an App Inventor project.
    """

    def __init__(self, filename):
        if not isinstance(filename, str) or (not isdir(filename) and filename[-4:] == '.aia'):
            self.zipfile = ZipFile(filename)
        else:
            self.zipfile = None

        self.assets = []
        """
        A list of assets contained in the project.

        :type: list[aiatools.aia.AIAAsset]
        """

        self.filename = filename
        """
        The filename or file-like that is the source of the project.

        :type: basestring or file
        """

        self.properties = {}
        """
        The contents of the project.properties file.

        :type: Properties
        """

        self._screens = NamedCollection()
        self.screens = Selector(self._screens)
        """
        A :py:class:`~aiatools.selectors.Selector` over the components of type
        :py:class:`~aiatools.component_types.Screen` in the project.

        For example, if you want to know how many screens are in a project run:

        .. doctest::

            >>> with AIAFile('test_aias/LondonCholeraMap.aia') as aia:
            ...     len(aia.screens)
            1

        :type: aiatools.selectors.Selector[aiatools.component_types.Screen]
        """

        self.components = UnionSelector(self._screens, 'components')
        """
        A :py:class:`~aiatools.selectors.Selector` over the component instances of all screens defined in the project.

        For example, if you want to know how many component instances are in a project run:

        .. doctest::

            >>> with AIAFile('test_aias/LondonCholeraMap.aia') as aia:
            ...     len(aia.components())  # Form, Map, Marker, Button
            4

        :type: aiatools.selectors.Selector[aiatools.common.Component]
        """

        self.blocks = UnionSelector(self._screens, 'blocks')
        """
        A :py:class:`~aiatoools.selectors.Selector` over the blocks of all screen defined in the project.

        For example, if you want to know how many blocks are in a project run:

        .. doctest::

            >>> with AIAFile('test_aias/LondonCholeraMap.aia') as aia:
            ...     len(aia.blocks())
            23

        :type: aiatools.selectors.Selector[aiatools.common.Block]
        """

        if self.zipfile:
            self._process_zip()
        else:
            self._process_dir()

    def close(self):
        if self.zipfile:
            self.zipfile.close()
        self.zipfile = None

    def __enter__(self):
        if self.zipfile:
            self.zipfile.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.zipfile:
            self.zipfile.__exit__(exc_type, exc_val, exc_tb)

    def _listfiles(self):
        names = []
        for dirname, dirs, files in os.walk(self.filename):
            names.extend([join(dirname, f) for f in files])

        return names

    def _process_zip(self):
        """
        Processes the contents of an AIA file into Python objects for further operation.
        """
        self.assets = []
        for name in self.zipfile.namelist():
            if name.startswith('assets/'):
                self.assets.append(AIAAsset(self, name))
                # TODO(ewpatton): Need to load extension JSON to extend language model
            elif name.startswith('src/'):
                if name.endswith('.scm'):
                    name = name[:-4]
                    form = self.zipfile.open('%s.scm' % name, 'r')
                    blocks = self.zipfile.open('%s.bky' % name, 'r')
                    screen = Screen(form=form, blocks=blocks)
                    self._screens[screen.name] = screen
            elif name.endswith('project.properties'):
                with self.zipfile.open(name) as prop_file:
                    self.properties = jprops.load_properties(prop_file)
            else:
                log.warning('Ignoring file in AIA: %s' % name)

    def _process_dir(self):
        """
        Processes the contents of a directory as if it were an AIA file and converts the content into Python objects
        for further operation.
        """
        self.assets = []
        asset_path = join(self.filename, 'assets')
        src_path = join(self.filename, 'src')
        for name in self._listfiles():
            if name.startswith(asset_path):
                self.assets.append(AIAAsset(None, name))
                # TODO(ewpatton): Need to load extension JSON to extend language model
            elif name.startswith(src_path):
                if name.endswith('.scm'):
                    name = name[:-4]
                    with open('%s.scm' % name, 'r') as form, open('%s.bky' % name, 'r') as blocks:
                        screen = Screen(form=form, blocks=blocks)
                        self._screens[screen.name] = screen
            elif name.endswith('project.properties'):
                with open(name, 'r') as prop_file:
                    self.properties = jprops.load_properties(prop_file)
            else:
                log.warning('Ignoring file in directory: %s' % name)




# -*- mode: python; coding: utf-8; -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""
aiatools.algebra defines the expressions and evaluation rules for querying the contents of AIA files.
"""
import collections


__author__ = 'Evan W. Patton <ewpatton@mit.edu>'


def _reduce_expression(expr, op):
    while isinstance(expr, Expression) and not isinstance(expr, Atom):
        expr = expr(op)
    return expr


def identity(x):
    """
    Helper function that returns its input.

    >>> identity("string")
    'string'
    >>> identity(True)
    True
    >>> identity(None)

    :param x: any value
    :return: x
    """
    return x


def needs_eval(x):
    """
    Tests whether its input needs ot be evaluated.

    >>> needs_eval(Expression())
    True
    >>> needs_eval(Atom())
    False

    :param x: the input expression
    :type x: Expression|callable
    :return:
    """
    if isinstance(x, Expression):
        return not isinstance(x, Atom)
    elif isinstance(x, collections.Callable):
        return True
    return False


class Expression(object):
    """
    Base interface for constructing expressions over App Inventor projects.

    Expressions by default support the following operations:

    1. ``left == right``: Accepts an entity if and only if ``left(entity) == right(entity)``
    2. ``left != right``: Accepts an entity if and only if ``left(entity) != right(entity)``
    3. ``left < right``: Accepts an entity if and only if ``left(entity) < right(entity)``
    4. ``left > right``: Accepts an entity if and only if ``left(entity) > right(entity)``
    5. ``left <= right``: Accepts an entity if and only if ``left(entity) <= right(entity)``
    6. ``left >= righ``: Accepts an entity if and only if ``left(entity >= right(entity)``
    7. ``left & right``: Accepts an entity if and only if ``left(entity) and right(entity)`` is True.
    8. ``left | right``: Accepts an entity if and only if ``left(entity) or right(entity)`` is True.
    9. ``~expr``: Accepts an entity if and only if ``not expr(entity)`` is True
    """
    def __eq__(self, other):
        """
        Constructs a new EquivalenceExpression with this expression as the left hand side and ``other`` as the right
        hand side.

        :param other: The other thing to check for equivalence with this expression when evaluated.
        :type other: Expression
        :return: A new equivalence expression
        :rtype: EquivalenceExpression
        """
        return EquivalenceExpression(self, other)

    def __ne__(self, other):
        """
        Constructs a new NotnequivalenceExpression with this expression as the left hand side and ``expr`` as the right
        hand side.

        :param other: The other expression to check for nonequivalence with this expression when evaluated.
        :type other: Expression
        :return:
        """
        return NonequivalenceExpression(self, other)

    def __hash__(self):
        return id(self)

    def __lt__(self, other):
        return LessThanExpression(self, other)

    def __gt__(self, other):
        return GreaterThanExpression(self, other)

    def __le__(self, other):
        return LessThanOrEqualExpression(self, other)

    def __ge__(self, other):
        return GreaterThanOrEqualExpression(self, other)

    def __and__(self, other):
        return AndExpression(self, other)

    def __or__(self, other):
        return OrExpression(self, other)

    def __invert__(self):
        return NotExpression(self)

    def __call__(self, operand, *args, **kwargs):
        raise NotImplementedError


class BinaryExpression(Expression):
    """
    Abstract base class for an Expression taking two clauses.

    Concrete implementations of this class must provide an implementation of the ``__call__`` special method to evaluate
    the truth value of the left and right hand sides.

    Parameters
    ----------
    left : Expression
        The left hand side of the binary expression.
    right : Expression
        The right hand side of the binary expression.
    """
    def __init__(self, left, right):
        self.left = ComputedAttribute(left) if isinstance(left, collections.Callable) and not isinstance(left, Expression) else left
        self.right = ComputedAttribute(right) if isinstance(right, collections.Callable) and not isinstance(right, Expression) else right

    def __call__(self, operand, *args, **kwargs):
        raise NotImplementedError


class EquivalenceExpression(BinaryExpression):
    """
    :py:class:`EquivalenceExpression` compares the output of two expressions for equivalent values, however == is
    defined on those values. An EquivalenceExpression is typically constructed by using the == operator on an existing
    pair of expressions, for example:

        >>> from aiatools.attributes import name
        >>> name == 'Button1'
        NamedAttributeTuple(('name', 'instance_name')) == 'Button1'
    """
    def __call__(self, operand, *args, **kwargs):
        left_val = _reduce_expression(self.left, operand)
        right_val = _reduce_expression(self.right, operand)
        return (left_val == right_val) is True

    def __repr__(self):
        return '%r == %r' % (self.left, self.right)


class NonequivalenceExpression(BinaryExpression):
    """
    :py:class:`NonequivalenceExpression` compares the output of two expressions for nonequivalent values, however != is
    defined on those values. A NonequivalenceExpression is typically constructed by using the != operator on an existing
    pair of expressions, for example:

        >>> from aiatools.attributes import name
        >>> name != 'Button1'
        NamedAttributeTuple(('name', 'instance_name')) != 'Button1'
    """
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return (left_val != right_val) is True

    def __repr__(self):
        return '%r != %r' % (self.left, self.right)


class LessThanExpression(BinaryExpression):
    """
    :py:class:`LessThanExpression` compares the output of two expressions and returns True if the value of the left hand
    side of the expression is less than the value of the right hand expression, for the definition of the less than
    operation on the two values. A LessThanExpression is typically constructed by using the < operator on an existing
    pair of expressions, for example:

        >>> from aiatools.attributes import version
        >>> version < 5
        NamedAttribute('version') < 5
    """
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return (left_val < right_val) is True

    def __repr__(self):
        return '%r < %r' % (self.left, self.right)


class GreaterThanExpression(BinaryExpression):
    """
    :py:class:`GreaterThanExpression` compares the output of two expressions and returns True if the value of the left
    handl side of the expression is greater than the value of the right hand expression, for the definition of the
    greater than operation on the two values. A GreaterThanExpression is typically constructed by using the > operator
    on an existing pair of expressions, for example:

        >>> from aiatools.attributes import version
        >>> version > 5
        NamedAttribute('version') > 5
    """
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return (left_val > right_val) is True

    def __repr__(self):
        return '%r > %r' % (self.left, self.right)


class LessThanOrEqualExpression(BinaryExpression):
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return (left_val <= right_val) is True

    def __repr__(self):
        return '%r <= %r' % (self.left, self.right)


class GreaterThanOrEqualExpression(BinaryExpression):
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return (left_val >= right_val) is True

    def __repr__(self):
        return '%r >= %r' % (self.left, self.right)


class AndExpression(BinaryExpression):
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return left_val and right_val

    def __repr__(self):
        return '%r & %r' % (self.left, self.right)


class OrExpression(BinaryExpression):
    """

    """
    def __call__(self, operand, *args, **kwargs):
        left_val = self.left(operand) if isinstance(self.left, Expression) else self.left
        right_val = self.right(operand) if isinstance(self.right, Expression) else self.right
        return left_val or right_val

    def __repr__(self):
        return '%s | %s' % (self.left, self.right)


class NotExpression(Expression):
    """
    :py:class:`NotExpression` is a unary expression that logically negates the output of the expression it encapsulates.
    NotExpressions are typically instantiated by using the unary prefix operator ~ to invert the expression. Note that
    ~ binds tightly, so most expressions, unless they are :py:class:`Atom`, must be wrapped in parentheses.

    Note
    ----
    NotExpression will optimize its own inversion so that two operators will cancel one another out. For example:

        >>> from aiatools.attributes import disabled
        >>> ~disabled
        ~NamedAttribute('disabled')
        >>> ~~disabled
        NamedAttribute('disabled')

    Parameters
    ----------
    expr : Expression
        The expression to negate.
    """
    def __init__(self, expr):
        self.expr = ComputedAttribute(expr) if isinstance(expr, collections.Callable) and not isinstance(expr, Expression) else expr

    def __call__(self, operand, *args, **kwargs):
        return not (self.expr(operand) if isinstance(self.expr, Expression) else self.expr)

    def __repr__(self):
        if isinstance(self.expr, Expression) and not isinstance(self.expr, (Functor, Atom)):
            return '~(%r)' % self.expr
        return '~%r' % self.expr

    def __invert__(self):
        return self.expr


class Atom(Expression):
    """
    :py:class:`Atom` represents an entity in the grammar, such as a specific component type (Button) or block type
    (component_set_get). Atoms cannot be modified and evaluate to themselves.
    """
    def __eq__(self, other):
        if isinstance(other, Atom):
            return other is self
        elif isinstance(other, Expression):
            return other == self
        elif isinstance(other, str):
            return other == self()
        else:
            return False

    def __hash__(self):
        return id(self)

    def __ne__(self, other):
        if isinstance(other, Atom):
            return other is not self
        else:
            return other != self

    def __call__(self, operand, *args, **kwargs):
        return self


class Functor(Expression):
    """
    :py:class:`Functor` is an abstract base class that serves as the root of the class tree of classes that apply
    functions to entities. Unlike most expressions, these typically compute non-Boolean values that may then undergo
    further computation.
    """
    def __call__(self, obj, *args, **kwargs):
        if needs_eval(obj):
            return FunctionComposition(self, obj)
        raise NotImplemented()


class Collection(Atom):
    """
    :py:class:`Collection` is an :py:class:`Atom` wrapping a collection, such as a list or tuple, of entities. Calls to
    the Collection will filter the collection given an Expression.

    Parameters
    ----------
    collection : collections.Iterable[aiatools.common.Component|aiatools.common.Block]
        The Python collection of entities to be wrapped into the new atomic collection.
    """
    def __init__(self, collection):
        self.collection = collection

    def __call__(self, *args, **kwargs):
        return Collection(list(filter(args[0], self.collection)))


class FunctionComposition(Functor):
    """
    :py:class:`ComposedAttribute` is a :py:class:`Functor` that wraps other Functors, Python functions, or lambda
    expressions. Functions are evaluated from left to right.

        >>> from aiatools import *
        >>> isinstance(root_block(declaration), FunctionComposition)
        True
        >>> proctest = AIAFile('test_aias/ProcedureTest2.aia')
        >>> proctest.blocks(is_procedure).callers(root_block(declaration)).map(fields.PROCNAME)
        ['i_am_called']

    Parameters
    ----------
    *args
        Functions to compose into a new function
    """
    def __init__(self, *args):
        self.functors = list(args) + [identity]

    def __call__(self, obj, *args, **kwargs):
        if needs_eval(obj):
            self.functors = [obj] + self.functors
            return self
        else:
            for op in self.functors:
                obj = op(obj)
            return obj


class ComputedAttribute(Functor):
    """
    :py:class:`ComputedAttribute` is a :py:class:`Functor` that wraps a Python function or lambda expression. This
    allows for arbitrary computations to be used in evaluating entities in a project.

    Parameters
    ----------
    functor : callback
        The functor that should be applied to the entity when the ComputedAttribute needs to be computed. Note that the
        return value is not memoized, so the given functor should be time efficient as possible when computing its
        value.
    """
    def __init__(self, functor):
        self.functor = functor

    def __call__(self, obj, *args, **kwargs):
        if needs_eval(obj):
            return FunctionComposition(self, obj)
        return self.functor(obj, *args, **kwargs)

    def __hash__(self):
        return hash(self.functor)

    def equals(self, other):
        return id(self.functor) == id(other.functor)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.functor)


and_ = AndExpression
or_ = OrExpression
not_ = NotExpression



# -*- mode: python; coding: utf-8 -*-
# Copyright © 2017 Massachusetts Institute of Technology, All rights reserved.

"""
The :py:mod:`aiatools.component_types` module defines the components for App Inventor. Component types are
programmatically constructed during module compilation from the simple_components.json used to populate App Inventor's
online development environment.
"""

import json
import xml.etree.ElementTree as ETree

import pkg_resources

from aiatools.selectors import Selector, NamedCollection, Selectors
from .common import *

__author__ = 'Evan W. Patton <ewpatton@mit.edu>'


# noinspection PyShadowingBuiltins
class ComponentContainer(Component, Selectors):
    """
    :py:class:`ComponentContainer` models the App Inventor ComponentContainer class.

    Parameters:
        parent (ComponentContainer, optional): The parent of the container. May be None for :py:class:`.Screen`.
        uuid (basestring): The UUID for the component container.
        type (str): The type of the component container.
        name (basestring): The name of the component container.
        version (str): The version number of the component container's type at the time the project was last saved.
        properties (dict[basestring, T], optional): The properties of the component as a dictionary. The values are
                                                    dependent on the key (property).
        components (list[Component], optional): A list of components contained by the container.
    """
    def __init__(self, parent, uuid, type, name, version, properties=None, components=None):
        super(ComponentContainer, self).__init__(parent, uuid, type, name, version, properties)
        self._children = [] if components is None else list(components)

    def __iter__(self):
        """
        Iterate over the children of the container.

        :return: An iterator over the container's children.
        :rtype: collections.Iterable[Component]
        """
        return iter(self._children)

    def itervalues(self):
        """
        Iterate over the values of this container. For :py:class:`ComponentContainer`, the values are the children
        of the container.

        :return: Iterator over the container's children.
        :rtype: collections.Iterable[Component]
        """
        return iter(self)

    def children(self):
        """
        Iterate over the child components in the container

        :return: An iterator over the components in the container.
        :rtype: collections.Iterable[Component]

        .. versionadded:: 0.1
        """
        return self._children

    def _components(self, *args, **kwargs):
        items = {item.id: item for item in RecursiveIterator(self)}
        items[self.id] = self
        return Selector(NamedCollection(items)).components(*args, **kwargs)

    @classmethod
    def from_json(cls, parent, json_repr):
        """
        Constructs a :py:class:`ComponentContainer`

        :param parent: The parent container of the new container.
        :type parent: ComponentContainer
        :param json_repr: The JSON representation of the component from the Screen definition.
        :type json_repr: dict[str, T]
        :return: A newly constructed ComponentContainer from the JSON representation.
        :rtype: ComponentContainer
        """
        typename = json_repr['$Type']
        if typename in globals():
            type = globals()[typename]
        else:
            type = Extension(typename)
        properties = {k: v for k, v in json_repr.items() if k not in Component._DISALLOWED_KEYS}
        container = cls(parent, json_repr['Uuid'], type, json_repr['$Name'], json_repr['$Version'], properties)
        for component_description in json_repr['$Components']:
            if '$Components' in component_description:
                child = ComponentContainer.from_json(container, component_description)
            else:
                child = Component.from_json(container, component_description)
            container._children.append(child)
        return container

    components = property(_components, doc="""
    Returns a :py:class:`~aiatools.selectors.Selector` over the components in the container.
    
    :type: Selector[Component]
    """)


class Screen(ComponentContainer):
    """
    The :py:class:`Screen` class provides a Python representation of an App Inventor Screen.

    The Screen object encapsulates both its descendant components and the blocks code prescribing the behavior of the
    Screen's contents.

    Parameters
    ----------
    name : basestring, optional
        The name of the screen.
    components : list[Component], optional
        A list of immediate components that are the children of the screen.
    form : string | file, optional
        A pathanme, string, or file-like that contains a Screen's Scheme (.scm) file.
    blocks : string | file, optional
        A pathname, string, or file-like that contains a Screen's Blocks (.bky) file.
    """
    def __init__(self, name=None, components=None, form=None, blocks=None):
        self.uuid = 0
        self.name = name
        self.path = name
        self._children = components
        self._blocks = NamedCollection()
        self.ya_version = None
        self.blocks_version = None
        if form is not None:
            if isinstance(form, str):
                form_json = json.loads(form)
            else:
                form_contents = [line.decode('utf-8') if hasattr(line, 'decode') else line for line in form.readlines()]
                if form_contents[1] != '$JSON\n' and form_contents[1] != b'$JSON\n':
                    raise RuntimeError('Unknown Screen format: %s' % form_contents[1])
                form_json = json.loads(form_contents[2])

            self.name = name or form_json['Properties']['$Name']
            super(Screen, self).__init__(parent=None,
                                         uuid=form_json['Properties']['Uuid'],
                                         type=Form,
                                         name=self.name,
                                         version=form_json['Properties']['$Version'],
                                         components=components)
            self._process_components_json(form_json['Properties']['$Components']
                                          if '$Components' in form_json['Properties'] else [])
            self.properties = {
                key: value for key, value in form_json.items() if key not in Component._DISALLOWED_KEYS
            }
            self.ya_version = int(form_json['YaVersion'])
        else:
            super(Screen, self).__init__(None, '0', Form, name or 'Screen1', '20', components=components)
        self.id = self.name
        if isinstance(blocks, str):
            blocks_content = blocks
        else:
            blocks_content = None if blocks is None else blocks.read()
        xml_root = None if blocks is None else ETree.fromstring(blocks_content)
        if xml_root is not None:
            for child in xml_root:
                if child.tag.endswith('yacodeblocks'):
                    self.blocks_version = int(child.attrib['language-version'])
                    self.ya_version = max(self.ya_version, int(child.attrib['ya-version']))
                else:
                    block = Block.from_xml(self, child, self.blocks_version)
                    self._blocks[block.id] = block
        self.blocks = Selector(self._blocks)

    def _process_components_json(self, components):
        self._children = []
        for component_description in components:
            if '$Components' in component_description:  # component container
                component = ComponentContainer.from_json(self, component_description)
            else:
                component = Component.from_json(self, component_description)
            self._children.append(component)

    def __iter__(self):
        for child in RecursiveIterator(self):
            yield child
        for child in self._blocks.values():
            yield child

    def __repr__(self):
        return "Screen(%s)" % repr(self.name)

    def __str__(self):
        return self.name


def list_to_dict(iterable, key='name'):
    return {i[key]: i for i in iterable}


def _load_component_types():
    """
    Loads the descriptions of App Inventor components from simple_components.json and populates the module with
    instances of ComponentType for each known type.
    """
    with open(pkg_resources.resource_filename('aiatools', 'simple_components.json')) as _f:
        _component_descriptors = json.load(_f)
        for _descriptor in _component_descriptors:
            _methods = list_to_dict(_descriptor['methods'])
            _events = list_to_dict(_descriptor['events'])
            _properties = list_to_dict(_descriptor['properties'])
            for _prop in _descriptor['blockProperties']:
                if _prop['name'] in _properties:
                    _properties[_prop['name']].update(_prop)
                else:
                    _properties[_prop['name']] = _prop
                    _prop['editorType'] = None
                    _prop['defaultValue'] = None
            _component = Screen if _descriptor == 'Form' else \
                ComponentType(_descriptor['name'], _methods, _events, _properties)
            _component.type = _descriptor['type']
            _component.external = bool(_descriptor['external'])
            _component.version = int(_descriptor['version'])
            _component.category_string = _descriptor['categoryString']
            _component.help_string = _descriptor['helpString']
            _component.show_on_palette = bool(_descriptor['showOnPalette'])
            _component.visible = not bool(_descriptor['nonVisible'])
            _component.icon_name = _descriptor['iconName']
            globals()[_component.name] = _component
            Component.TYPES[_component.name] = _component


_load_component_types()
